from typing import Dict, Any

import ccxt
from ccxt.base.types import Balances
import ccxt.pro as ccxtpro

'''
Why override fetch_balance?
    balances['total'] empty
But you can find that from exchange raw response under balances['info']['balances'] (verbose=True).
'''
def _populate_balance_total_if_missing(
        balances : Dict[str, Any]
    ):
    for ccy_balance in balances['info']['balances']:
        ccy = ccy_balance['asset']
        free = float(ccy_balance.get('free', 0))
        locked = float(ccy_balance.get('locked', 0))
        total = free + locked
        if total!=0 and ccy not in balances['total']:
            balances['total'][ccy] = total
class Binance(ccxt.binance):
    def __init__(self, *args: Dict[str, Any]) -> None:
        super().__init__(*args) # type: ignore
    
    def fetch_balance(self, params={}) -> Balances: # type: ignore
        balances = super().fetch_balance(params=params)
        _populate_balance_total_if_missing(balances)
        return balances

class BinanceAsync(ccxtpro.binance):
    def __init__(self, *args: Dict[str, Any]) -> None:
        super().__init__(*args) # type: ignore

    async def fetch_balance(self, params={}) -> Balances: # type: ignore
        balances = await super().fetch_balance(params=params)
        _populate_balance_total_if_missing(balances)
        return balances


def is_int_string(s: str) -> bool:
    if not s:
        return False
    return s.lstrip('-+').isdigit()

def is_float_string(s: str) -> bool:
    if not s:
        return False
    try:
        float(s)
        return True
    except ValueError:
        return False



from datetime import datetime, timedelta
from typing import Dict

def parse_trading_window(
            today : datetime,
            window : Dict[str, str]
        ) :
        window_start : str = window['start']
        window_end : str = window['end']

        DayOfWeekMap : Dict[str, int] = {
            'Mon' : 0,
            'Tue' : 1,
            'Wed' : 2,
            'Thur' : 3,
            'Fri' : 4,
            'Sat' : 5,
            'Sun' : 6
        }
        today_dayofweek = today.weekday()

        window_start_dayofweek : int = DayOfWeekMap[window_start.split('_')[0]]
        window_start_hr : int = int(window_start.split('_')[-1].split(':')[0])
        window_start_min : int = int(window_start.split('_')[-1].split(':')[1])
        dt_window_start = today + timedelta(days=(window_start_dayofweek-today_dayofweek))
        dt_window_start = dt_window_start.replace(hour=window_start_hr, minute=window_start_min)

        window_end_dayofweek : int = DayOfWeekMap[window_end.split('_')[0]]
        window_end_hr : int = int(window_end.split('_')[-1].split(':')[0])
        window_end_min : int = int(window_end.split('_')[-1].split(':')[1])
        dt_window_end = today + timedelta(days=(window_end_dayofweek-today_dayofweek))
        dt_window_end = dt_window_end.replace(hour=window_end_hr, minute=window_end_min)

        return {
            'today' : today,
            'start' : dt_window_start,
            'end' : dt_window_end,
            'in_window' : (today<=dt_window_end) and (today>=dt_window_start)
        }


import unittest
from typing import List, Dict, Union
import time

from numpy import equal

from util.simple_math import generate_rand_nums, round_to_level, compute_adjacent_levels, bucket_series, bucketize_val, msd_power, round_to_sigfigs

class SimpleMathTests(unittest.TestCase):

    def test_generate_rand_nums(self):
        range_min : float = 0
        range_max : float = 1
        size : int = 100
        percentage_in_range : float = 91
        abs_min : float = -0.5
        abs_max : float = 1.1

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )
        
        assert(len(rand_nums)==size)
        assert(len([x for x in rand_nums if x>=range_min and x<=range_max]) == (percentage_in_range/100) * size)
        assert(len([x for x in rand_nums if x<abs_min or x>abs_max]) == 0)


        range_min = -1
        range_max  = 1
        percentage_in_range = 91
        abs_min = -1.5
        abs_max = 1.5

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )
        
        assert(len(rand_nums)==size)
        assert(len([x for x in rand_nums if x>=range_min and x<=range_max]) == (percentage_in_range/100) * size)
        assert(len([x for x in rand_nums if x<abs_min or x>abs_max]) == 0)


        range_min = 0
        range_max  = 100
        percentage_in_range = 91
        abs_min = -150
        abs_max = 150

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )
        
        assert(len(rand_nums)==size)
        assert(len([x for x in rand_nums if x>=range_min and x<=range_max]) == (percentage_in_range/100) * size)
        assert(len([x for x in rand_nums if x<abs_min or x>abs_max]) == 0)


        range_min = -100
        range_max  = 100
        percentage_in_range = 91
        abs_min = -150
        abs_max = 150

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )
        
        assert(len(rand_nums)==size)
        assert(len([x for x in rand_nums if x>=range_min and x<=range_max]) == (percentage_in_range/100) * size)
        assert(len([x for x in rand_nums if x<abs_min or x>abs_max]) == 0)

    def test_round_to_level(self):
        prices = [ 
            { 'price' : 15080, 'rounded' : 15000}, 
            { 'price' : 15180, 'rounded' : 15200}, 
            { 'price' : 25080, 'rounded' : 25200}, 
            { 'price' : 25180, 'rounded' : 25200}, 
            { 'price' : 25380, 'rounded' : 25500}, 
            { 'price' : 95332, 'rounded' : 95000}, 
            { 'price' : 95878, 'rounded' : 96000}, 
            { 'price' : 103499, 'rounded' : 103000}, 
            { 'price' : 103500, 'rounded' : 104000}, 
            { 'price' : 150800, 'rounded' : 150000}, 
            { 'price' : 151800, 'rounded' : 152000}
        ]
        for entry in prices:
            price = entry['price']
            expected = entry['rounded']
            rounded_price = round_to_level(price, level_granularity=0.01)
            print(f"{price} rounded to: {rounded_price}")
            assert(rounded_price==expected)

    def test_compute_adjacent_levels(self):
        # ******* non-zero level_granularity *******
        gold_price = 4450
        level_granularity = 0.025 # So levels are $100 apart
        adjacent_levels = compute_adjacent_levels(num=gold_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [4100,4200,4300,4400,4500,4600,4700])

        btc_price = 95000
        level_granularity = 0.01 # So levels are $1000 apart
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [92000,93000,94000,95000,96000,97000,98000])

    def test_compute_adjacent_levels_w_zero_level_granularity(self):
        
        # ******* zero level_granularity *******
        level_granularity = 0

        btc_price = 48000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [20000, 30000, 40000, 50000, 60000, 70000, 80000])

        btc_price = 88000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [60000, 70000, 80000, 90000, 100000, 110000, 120000])

        btc_price = 118000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [90000, 100000, 110000, 120000, 130000, 140000, 150000])

        btc_price = 178000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [150000, 160000, 170000, 180000, 190000, 200000, 210000])

        btc_price = 378000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [100000, 200000, 300000, 400000, 500000, 600000, 700000])

        sol_price = 18
        adjacent_levels = compute_adjacent_levels(num=sol_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [15, 16, 17, 18, 19, 20, 21])

        sol_price = 78
        adjacent_levels = compute_adjacent_levels(num=sol_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [50, 60, 70, 80, 90, 100, 110])

        sol_price = 108
        adjacent_levels = compute_adjacent_levels(num=sol_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [80, 90, 100, 110, 120, 130, 140])

        sol_price = 408
        adjacent_levels = compute_adjacent_levels(num=sol_price, level_granularity=level_granularity, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [100, 200, 300, 400, 500, 600, 700])

    
    def test_compute_adjacent_levels_w_zero_level_granularity_and_msd_adj(self):
        
        # ******* zero level_granularity and msd_adj *******
        level_granularity = 0
        msd_adj = -1

        btc_price = 68000
        adjacent_levels = compute_adjacent_levels(num=btc_price, level_granularity=level_granularity, msd_adj=msd_adj, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [65000, 66000, 67000, 68000, 69000, 70000, 71000])

        sol_price = 78
        adjacent_levels = compute_adjacent_levels(num=sol_price, level_granularity=level_granularity, msd_adj=msd_adj, num_levels_per_side=3)
        assert(adjacent_levels)
        assert(len(adjacent_levels)==7)
        equal(adjacent_levels, [75, 76, 77, 78, 79, 80, 81])
        

    def test_bucket_series(self):

        level_granularity : float = 0.1

        range_min : float = 0
        range_max : float = 1
        size : int = 100
        percentage_in_range : float = 91
        abs_min : float = -0.5
        abs_max : float = 1.1

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )

        buckets : Dict[
            str, 
            Dict[str,Union[float, List[float]]]
        ] = bucket_series(
                                                values = rand_nums,
                                                outlier_threshold_percent = 10,
                                                level_granularity=level_granularity
                                            )
        
        bucketized = []
        for num in rand_nums:
            bucketized.append(
                bucketize_val(num, buckets=buckets)
            )

        
        range_min = -1
        range_max = 1
        size : int = 100
        percentage_in_range = 91
        abs_min = -1.5
        abs_max = 1.5

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )

        buckets = bucket_series(
                                    values = rand_nums,
                                    outlier_threshold_percent = 10,
                                    level_granularity=level_granularity
                                )
        

        range_min = 0
        range_max = 100
        size : int = 100
        percentage_in_range = 91
        abs_min = -0.5
        abs_max = 150

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )

        buckets = bucket_series(
                                    values = rand_nums,
                                    outlier_threshold_percent = 10,
                                    level_granularity=level_granularity
                                )
        

        range_min = -100
        range_max = 100
        size : int = 100
        percentage_in_range = 91
        abs_min = -150
        abs_max = 150

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )

        buckets = bucket_series(
                                    values = rand_nums,
                                    outlier_threshold_percent = 10,
                                    level_granularity=level_granularity
                                )
        

        range_min = 20_000
        range_max = 120_000
        size : int = 100
        percentage_in_range = 91
        abs_min = 15_000
        abs_max = 130_000

        rand_nums : List[float] = generate_rand_nums(
                                        range_min = range_min,
                                        range_max = range_max,
                                        size = size, 
                                        percent_in_range = percentage_in_range,
                                        abs_min = abs_min,
                                        abs_max = abs_max
                                    )

        buckets = bucket_series(
                                    values = rand_nums,
                                    outlier_threshold_percent = 10,
                                    level_granularity=level_granularity
                                )

    def test_msd_power(self):
        # Integers
        self.assertEqual(msd_power(34523), 4)
        self.assertEqual(msd_power(99999), 4)
        self.assertEqual(msd_power(100000), 5)
        self.assertEqual(msd_power(999999999), 8)
        self.assertEqual(msd_power(1), 0)
        self.assertEqual(msd_power(9), 0)
        self.assertEqual(msd_power(10), 1)

        # Decimals > 1
        self.assertEqual(msd_power(3452.3), 3)
        self.assertEqual(msd_power(12.345), 1)
        self.assertEqual(msd_power(9.87654321), 0)

        # Numbers between 0 and 1
        self.assertEqual(msd_power(0.056), -2)      # 5.6 × 10⁻²
        self.assertEqual(msd_power(0.0056), -3)     # 5.6 × 10⁻³
        self.assertEqual(msd_power(0.000999), -4)
        self.assertEqual(msd_power(0.1), -1)
        self.assertEqual(msd_power(0.999), -1)
        self.assertEqual(msd_power(0.000000123), -7)

        # Negative numbers (should ignore sign)
        self.assertEqual(msd_power(-123.5), 2)
        self.assertEqual(msd_power(-0.00456), -3)

        # Special / edge cases
        self.assertIsNone(msd_power(0))
        self.assertIsNone(msd_power(0.0))
        # Very large / very small (log10 should handle)
        self.assertEqual(msd_power(1e20), 20)
        self.assertEqual(msd_power(5.67e-15), -15)

    def test_round_to_sigfigs(self):
        # Using default sigfigs=6

        # Large numbers – integer part dominant
        self.assertEqual(round_to_sigfigs(34523.12345, sigfigs=6), 34523.1)
        self.assertEqual(round_to_sigfigs(34523.12345, sigfigs=7), 34523.12)
        self.assertEqual(round_to_sigfigs(123456.789), 123457)
        self.assertEqual(round_to_sigfigs(999999.5), 1000000)     # rounds up
        self.assertEqual(round_to_sigfigs(9876543210), 9876540000)

        # Numbers around 1–1000
        self.assertEqual(round_to_sigfigs(123.45678), 123.457)
        self.assertEqual(round_to_sigfigs(9.87654321), 9.87654)
        self.assertEqual(round_to_sigfigs(1.00000), 1.0)

        # Small numbers (decimal dominant)
        self.assertEqual(round_to_sigfigs(0.056789123), 0.0567891)

        # Very small scientific notation style
        self.assertEqual(round_to_sigfigs(1.23456789e-8), 1.23457e-8)
        self.assertEqual(round_to_sigfigs(9.87654321e-12), 9.87654e-12)

        # Negative numbers
        self.assertEqual(round_to_sigfigs(-34523.12345), -34523.1)
        self.assertEqual(round_to_sigfigs(-0.0056789), -0.0056789)

        # Edge cases
        self.assertEqual(round_to_sigfigs(0), 0.0)
        self.assertEqual(round_to_sigfigs(0.0), 0.0)

        # With explicit sigfigs parameter
        self.assertEqual(round_to_sigfigs(34523.12345, sigfigs=4), 34520)
        self.assertEqual(round_to_sigfigs(1234567, sigfigs=3), 1230000)

        '''
        Edge cases:
            round_to_sigfigs(0.0056789123) -> 0.0056789100000000006
            round_to_sigfigs(0.00034523123) -> 0.00034523100000000004
            round_to_sigfigs(0.056789, sigfigs=3) -> 0.056799999999999996
        These are extra tiny error from the repeated multiply/divide in float arithmetic with python.
        '''
        # self.assertEqual(round_to_sigfigs(0.0056789123), 0.00567891) # 0.0056789100000000006
        # self.assertEqual(round_to_sigfigs(0.00034523123), 0.000345231)
        # self.assertEqual(round_to_sigfigs(0.056789, sigfigs=3), 0.0568)

        NUM_CALLS = 10_000
        start = time.time()
        for _ in range(10_000):
            x = round_to_sigfigs(85.123456789, sigfigs=7)
        elapsed_ms = int((time.time() - start) *1000)
        avg_per_call_ms = elapsed_ms / NUM_CALLS
        print(f"elapsed_ms: {elapsed_ms} over {NUM_CALLS} calls, avg_per_call_ms: {avg_per_call_ms}.")
        assert(avg_per_call_ms<1)


from typing import Dict
import json
from datetime import datetime
import numpy as np

import logging

# What's this for? To avoid the following error when you json.dumps: TypeError: Object of type int64 is not JSON serializable
def recursive_clean_dict(data : Dict):
    logger = logging.getLogger()

    def _clean_val(v):
        if v is None:
            return None  # or whatever you prefer

        # NumPy integer types (int8/16/32/64, uint8/16/32/64)
        if isinstance(v, np.integer):
            return int(v)

        # NumPy boolean (very common escape)
        if isinstance(v, (np.bool_, bool)):  # bool just in case
            return bool(v)

        # NumPy floats
        if isinstance(v, (np.floating, np.float16, np.float32, np.float64)):
            return float(v)

        # NumPy datetime / timedelta (if you ever have them)
        if isinstance(v, np.datetime64):
            return np.datetime_as_string(v, unit='s')  # or your preferred format
        if isinstance(v, np.timedelta64):
            return str(v)  # or convert to seconds etc.

        # datetime objects (already good)
        if isinstance(v, datetime):
            return v.strftime("%Y%m%d %H:%M:%S")

        # catch anything else suspicious
        if isinstance(v, (np.void, np.str_, np.bytes_, np.generic)):
            return str(v)

        return v

    cleaned = {}
    for k, v in data.items():
        _v = _clean_val(v)
        
        if isinstance(_v, dict):
            cleaned[k] = recursive_clean_dict(v)

        elif isinstance(_v, list):
            l = []
            for x in _v:
                if isinstance(x, dict):
                    _x = recursive_clean_dict(x)
                else:
                    _x = _clean_val(x)
                l.append(_x)

            cleaned[k] = l
        
        else:
            cleaned[k] = _v

    return cleaned

    



import asyncio
import sys
import traceback
import os
import argparse
import json
import hashlib
import re
from datetime import datetime, timedelta, timezone
import time
import pytz
import arrow
from enum import Enum
import logging
import requests
from typing import Dict, Optional, Set, Any, Union, List
from redis import StrictRedis

from siglab_py.util.notification_util import dispatch_notification

current_filename = os.path.basename(__file__)

'''
google_monitor fetches messages from particular query. Then:
    a. Save (and accumulate) messages to message cache file (No duplicates) for further analysis.
        message_cache_file: str = f"google_search_messages.json"

    b. If any of keywords in message_keywords_filter matches words in message (--message_keywords_filter):
        - Publish to redis for strategy consumption, topic: param['mds']['topics']['google_alert']
        - Dispatch slack alert
        - If scripts runs on Windows, play a wav file (Feels free make modification play sounds on Ubuntu for example)

Usage:
    set PYTHONPATH=%PYTHONPATH%;D:\dev\siglab\siglab_py
    python google_monitor.py --apikey xxx --search_engine_id yyy --query "site:twitter.com @user_id1 @user_id2 some topic" --slack_info_url=https://hooks.slack.com/services/xxx --slack_critial_url=https://hooks.slack.com/services/xxx --slack_alert_url=https://hooks.slack.com/services/xxx

    alert_wav_path
        Point it to wav file for alert notification. It's using 'winsound', i.e. Windows only.
        Set to None otherwise.
        
Google API: https://console.cloud.google.com/apis/credentials?project=YOUR_PROJECT
		name: YOUR_API_KEY_NAME
		apikey: ?????

Google Search Engine
	To create
		name: siglab_py_search: https://programmablesearchengine.google.com/controlpanel/create
			<script async src="https://cse.google.com/cse.js?cx=YOUR_SEARCH_ENGINE_ID">
			</script>
			<div class="gcse-search"></div>
	Then enable it: https://console.developers.google.com/apis/api/customsearch.googleapis.com/overview?project=?????

launch.json for Debugging from VSCode:
    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python: Current File",
                "type": "python",
                "request": "launch",
                "program": "${file}",
                "console": "integratedTerminal",
                "justMyCode": false,
                "args" : [
                        "--apikey", "xxx",
                        "--search_engine_id", "yyy",
                        "--query", "site:twitter.com @user_id1 @user_id2 some topic",
                        "--slack_info_url", "https://hooks.slack.com/services/xxx",
                        "--slack_critial_url", "https://hooks.slack.com/services/xxx",
                        "--slack_alert_url", "https://hooks.slack.com/services/xxx",
                    ],
            }
        ]
    }
'''

param: Dict[str, Any] = {
    'apikey': os.getenv('GOOGLE_APIKEY', 'xxx'),
    'search_engine_id': os.getenv('GOOGLE_SEARCH_ENGINE_ID', 'xxx'),
    'num_results' : 10,
    'query' : '',
    'alert_wav_path' : r"d:\sounds\terrible.wav",
    "num_shouts" : 5, # How many times 'alert_wav_path' is played
    "loop_freq_ms" : 1000*60*15, # Google allow max 100 calls per day free.
    'current_filename' : current_filename,

    'notification' : {
        'footer' : None,

        # slack webhook url's for notifications
        'slack' : {
            'info' : { 'webhook_url' : None },
            'critical' : { 'webhook_url' : None },
            'alert' : { 'webhook_url' : None },
        }
    },
    
    'mds': {
        'topics': {
            'tg_alert': 'tg_alert'
        },
        'redis': {
            'host': 'localhost',
            'port': 6379,
            'db': 0,
            'ttl_ms': 1000 * 60 * 15
        }
    }
}

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

logging.Formatter.converter = time.gmtime
logger: logging.Logger = logging.getLogger()
log_level: int = logging.INFO
logger.setLevel(log_level)
format_str: str = '%(asctime)s %(message)s'
formatter: logging.Formatter = logging.Formatter(format_str)
sh: logging.StreamHandler = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)

def log(message: str, log_level: LogLevel = LogLevel.INFO) -> None:
    if log_level.value < LogLevel.WARNING.value:
        logger.info(f"{datetime.now()} {message}")
    elif log_level.value == LogLevel.WARNING.value:
        logger.warning(f"{datetime.now()} {message}")
    elif log_level.value == LogLevel.ERROR.value:
        logger.error(f"{datetime.now()} {message}")

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore
    parser.add_argument("--apikey", help="API key", default=None)
    parser.add_argument("--search_engine_id", help="Google search engine ID", default=None)
    parser.add_argument("--num_results", help="Max number items to fetch", default=10)
    parser.add_argument("--query", help="Query - what are you looking for?", default=None)
    parser.add_argument("--slack_info_url", help="Slack webhook url for INFO", default=None)
    parser.add_argument("--slack_critial_url", help="Slack webhook url for CRITICAL", default=None)
    parser.add_argument("--slack_alert_url", help="Slack webhook url for ALERT", default=None)

    args = parser.parse_args()
    
    param['apikey'] = args.apikey
    param['search_engine_id'] = args.search_engine_id
    param['num_results'] = args.num_results
    param['query'] = args.query

    param['notification']['slack']['info']['webhook_url'] = args.slack_info_url
    param['notification']['slack']['critical']['webhook_url'] = args.slack_critial_url
    param['notification']['slack']['alert']['webhook_url'] = args.slack_alert_url

    param['notification']['footer'] = f"From {param['current_filename']}"

    print(f"Startup args: {args}") # Dont use logger, not yet setup yet.
    print(f"param: {print(json.dumps(param, indent=2))}")

def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = param['mds']['redis']['host'],
                    port = param['mds']['redis']['port'],
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
        log(f"Failed to init redis connection. Will skip publishes to redis. {err_msg}")
        redis_client = None # type: ignore
    
    return redis_client

def search_google_custom(query, api_key, search_engine_id, num_results=10):
    url = 'https://www.googleapis.com/customsearch/v1'
    params = {
        'key': api_key,
        'cx': search_engine_id,
        'q': query,
        'num': num_results,
        'sort': 'date',
        'dateRestrict': 'd1'  # Restrict to most recent (adjust as needed: d1=day, m1=month, etc.)
    }
    
    response = requests.get(url, params=params)
    
    if response.status_code == 200:
        return response.json()
    else:
        log(f"Query error: {response.status_code} - {response.text}")
        return None
    
async def main() -> None:
    parse_args()

    message_cache_file: str = f"google_search_messages.json"
    log(f"message_cache_file: {message_cache_file}")

    notification_params : Dict[str, Any] = param['notification']

    processed_messages : List[Dict[str, Any]] = []
    seen_hashes : Set[str] = set()
    if os.path.exists(message_cache_file):
        with open(message_cache_file, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for line in lines:
                message_data = json.loads(line)
                message_hash: str = hashlib.sha256(message_data['message'].encode('utf-8')).hexdigest()

                message_data['datetime'] = pytz.UTC.localize(arrow.get(message_data['datetime']).datetime.replace(tzinfo=None))

                if message_hash not in seen_hashes:
                    seen_hashes.add(message_hash)
                    processed_messages.append(message_data)

            processed_messages = sorted(processed_messages, key=lambda m: m['datetime'])
    
    try:        
        redis_client: Optional[StrictRedis] = init_redis_client()
    except Exception as redis_err:
        redis_client = None
        log(f"Failed to connect to redis. Still run but not publishing to it. {redis_err}")

    while True:
        try:
            results = search_google_custom(param['query'], param['apikey'], param['search_engine_id'], param['num_results'])

            if results:
                if 'items' in results:
                    for item in results['items']:
                        title = item.get('title', 'No title')
                        snippet = item.get('snippet', 'No snippet')
                        link = item.get('link', 'No link')
                        published_date = item.get('pagemap', {}).get('metatags', [{}])[0].get('article:published_time', 'No date')

                        dt_message = datetime.now()
                        pattern = r'^\d+\s*(?:days?|day?|hours?|hour?|minutes?|minute?|seconds?|second?|h|m|s)\s*(?:ago)?.*?([A-Za-z]+\s+\d+,\s+\d{4},\s+\d+:\d+\s+[AP]M\s+ET)'
                        match = re.match(pattern, snippet)
                        if published_date == 'No date' and match:
                            published_date = match.group(1)
                            dt_message = datetime.strptime(published_date, '%b %d, %Y, %I:%M %p ET')
                        
                        snippet = re.sub(pattern, '', snippet).strip()
                        
                        timestamp_ms  = int(dt_message.timestamp() * 1000)
                        message_data: Dict[str, Any] = {
                            "timestamp_ms": timestamp_ms,
                            "datetime": dt_message.isoformat(), # Always in UTC
                            "title" : title,
                            "message": snippet,
                            "url" : link
                        }
                        json_str: str = json.dumps(message_data, ensure_ascii=False, sort_keys=True)
                        message_hash: str = hashlib.sha256(snippet.encode('utf-8')).hexdigest()
                        if (message_hash not in seen_hashes):
                            seen_hashes.add(message_hash)
                            processed_messages.append(message_data)

                            log(f"{message_data}")

                            dispatch_notification(title=f"{param['current_filename']} Incoming! {title}", message=message_data, footer=param['notification']['footer'], params=notification_params, log_level=LogLevel.CRITICAL, logger=logger) # type: ignore

                            with open(message_cache_file, 'a', encoding='utf-8') as f:
                                json.dump(message_data, f, ensure_ascii=False)
                                f.write('\n')

                            if param['alert_wav_path']and sys.platform == 'win32':
                                import winsound
                                for _ in range(param['num_shouts']):
                                    winsound.PlaySound(param['alert_wav_path'], winsound.SND_FILENAME)

                            if redis_client:
                                try:
                                    publish_topic = f"google_search"
                                    redis_client.publish(publish_topic, json_str)
                                    redis_client.setex(message_hash, param['mds']['redis']['ttl_ms'] // 1000, json_str)
                                    log(f"Published message {json_str} to Redis topic {publish_topic}", LogLevel.INFO)
                                except Exception as e:
                                    log(f"Failed to publish to Redis: {str(e)}", LogLevel.ERROR)

                await asyncio.sleep(int(param['loop_freq_ms'] / 1000))

            if processed_messages:
                oldest_message: Dict[str, Any] = min(processed_messages, key=lambda x: x['timestamp_ms'])
                newest_message: Dict[str, Any] = max(processed_messages, key=lambda x: x['timestamp_ms'])
                log(
                    json.dumps(
                        {
                            'num_messages': len(processed_messages),
                            'oldest': {
                                'timestamp_ms': oldest_message['timestamp_ms'],
                                'datetime': datetime.fromtimestamp(int(oldest_message['timestamp_ms']/1000),tz=timezone.utc).isoformat()
                            },
                            'latest': {
                                'timestamp_ms': newest_message['timestamp_ms'],
                                'datetime': datetime.fromtimestamp(int(newest_message['timestamp_ms']/1000),tz=timezone.utc).isoformat()
                            }
                        }, indent=2
                    ),
                    LogLevel.INFO
                )

        except Exception as e:
            log(f"Oops {str(e)} {str(sys.exc_info()[0])} {str(sys.exc_info()[1])} {traceback.format_exc()}", LogLevel.ERROR)
        finally:
            await asyncio.sleep(int(param['loop_freq_ms'] / 1000))

if __name__ == '__main__':
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        log("Stopped by user", LogLevel.INFO)
    except Exception as e:
        log(f"Unexpected error: {str(e)}", LogLevel.ERROR)


import math
from pprint import pformat
from typing import List, Dict, Any, Union
from abc import ABC, abstractmethod
import logging

from siglab_py.constants import OrderSide 
from siglab_py.algo.strategy_base import StrategyBase
from siglab_py.util.simple_math import compute_adjacent_levels

logger = logging.getLogger()
        
'''
Why OneShotStrategy when we have tp_algo? OneShotStrategy uses strategy_executor: we can use orderhist_cache_viewer.ipynb for post-trade analysis 
'''
class OneShotStrategy(StrategyBase):
    def __init__(self, *args: object) -> None:
        pass
    
    @staticmethod
    def order_notional_adj(
        algo_param : Dict,
    ) -> Dict[str, float]:
        return StrategyBase.order_notional_adj(algo_param) # type: ignore

    @staticmethod
    def allow_entry_initial(
        mid,
        lo_row_tm1,
        hi_row_tm1,
        last_candles,
        algo_param : Dict
    )  -> Dict[str, bool]:
        allow_long : bool = False
        allow_short : bool = False

        side = algo_param['side']
        entry_on_reversal = algo_param['entry_on_reversal'] if "entry_on_reversal" in algo_param else False
        target_entry_price = algo_param['target_entry_price'] if "target_entry_price" in algo_param else None

        normalized_ema_long_slope = lo_row_tm1['normalized_ema_long_slope']

        if side=='buy':
            allow_long = True
            if entry_on_reversal:
                allow_long = True if normalized_ema_long_slope>=0 else False
            if target_entry_price:
                if normalized_ema_long_slope<0:
                    # Prices still falling (Still waiting for reversal)
                    if mid>target_entry_price:
                        allow_long = False
                else:
                    # Reversal already happenned (Chasing longs)
                    if mid<target_entry_price:
                        allow_long = False

        elif side=='sell':
            allow_short = True
            if entry_on_reversal:
                allow_short = True if normalized_ema_long_slope<=0 else False                
            if target_entry_price:
                if normalized_ema_long_slope>0:
                    # Prices still rising (Still waiting for reversal)
                    if mid<target_entry_price:
                        allow_short = False
                else:
                    # Reversal already happenned (Chasing shorts)
                    if mid>target_entry_price:
                        allow_short = False

        allow_entry_initial_conditions  = {
            "side" : side,
            "entry_on_reversal" : entry_on_reversal,
            "target_entry_price" : target_entry_price,
            "lo_row_tm1['datetime']" : lo_row_tm1['datetime'],
            "last_candles[-1]['datetime']" : last_candles[-1]['datetime'],
            "normalized_ema_long_slope" : normalized_ema_long_slope
        }

        print(pformat(allow_entry_initial_conditions, indent=2, width=100))
            
        return {
            'long' : allow_long,
            'short' : allow_short
        }

    @staticmethod
    def allow_entry_final(
        mid, # Different from backtest but more appropriate
        lo_row_tm1,
        algo_param : Dict
    ) -> Dict[str, Union[bool, float, None]]:
        allow_long : bool = True
        allow_short : bool = True
        
        entry_price_long, entry_price_short = mid, mid

        tp_max_bps = algo_param['tp_max_percent']*100

        target_price_long : float = mid * (1 + tp_max_bps/10000)
        target_price_short : float = mid * (1 - tp_max_bps/10000)

        return {
                'long' : allow_long,
                'short' : allow_short,

                # In additional to allow or not, allow_entry_final also calculate a few things which you may need to mark the entry trades.
                'entry_price_long' : entry_price_long,
                'entry_price_short' : entry_price_short,
                'target_price_long' : target_price_long,
                'target_price_short' : target_price_short,
                'reference_price' : None
            }
    
    @staticmethod
    def sl_adj(
        max_unreal_live : float,
        pos_usdt : float,
        this_ticker_open_trades : List[Dict],
        algo_param : Dict,
        *args: Any, **kwargs: Any
    ) -> Dict[str, float]:
        running_sl_percent_hard = algo_param['sl_hard_percent']
        sl_adj_percent = algo_param['sl_adj_percent']

        # sl_hard_percent adj: Strategies where target_price not based on tp_max_percent, but variable
        max_pnl_potential_percent = None
        if any([ trade for trade in this_ticker_open_trades if 'target_price' in trade ]):
            max_pnl_potential_percent = max([ (trade['target_price']/trade['entry_price'] -1) *100 if trade['side']=='buy' else (trade['entry_price']/trade['target_price'] -1) *100 for trade in this_ticker_open_trades if 'target_price' in trade ])

            if max_pnl_potential_percent and max_pnl_potential_percent<algo_param['tp_max_percent']:
                tp_minmax_ratio = algo_param['tp_min_percent']/algo_param['tp_max_percent']
                tp_min_sl_hard_ratio  = algo_param['sl_hard_percent']/algo_param['tp_min_percent']

                tp_max_percent = max_pnl_potential_percent
                tp_min_percent = tp_minmax_ratio * tp_max_percent
                running_sl_percent_hard = tp_min_sl_hard_ratio * tp_min_percent
        
        max_pnl_percent_notional = max_unreal_live / pos_usdt * 100 if pos_usdt else 0

        if max_pnl_percent_notional>=algo_param['sl_hard_percent']:
            running_sl_percent_hard = min(
                    running_sl_percent_hard,
                    max(0, running_sl_percent_hard - max_pnl_percent_notional * sl_adj_percent)
                )

        assert(running_sl_percent_hard<=algo_param['sl_hard_percent'])

        return {
            'running_sl_percent_hard' : running_sl_percent_hard
        }
    
    @staticmethod
    def trailing_stop_threshold_eval(
        algo_param : Dict,
        *args: Any, **kwargs: Any
    ) -> Dict[str, float]:
        result = StrategyBase.trailing_stop_threshold_eval(algo_param)
        tp_min_percent = result['tp_min_percent']
        tp_max_percent = result['tp_max_percent']
        return {
            'tp_min_percent' : tp_min_percent,
            'tp_max_percent' : tp_max_percent
        }

    @staticmethod
    def tp_eval (
            mid : float,
            tp_max_target : float,
            pos_side : OrderSide
    ) -> bool:
        tp : bool = False
        if pos_side==OrderSide.BUY:
            tp = True if mid>=tp_max_target else False
        elif pos_side==OrderSide.SELL:
            tp = True if mid<=tp_max_target else False
        return tp

    @staticmethod
    def get_strategy_indicators() -> List[str]:
        return [ 
            'lo_row_tm1:close',
            'hi_row_tm1:ema_close',
            'lo_row_tm1:normalized_ema_long_slope',
            'lo_row_tm1:atr', 'lo_row_tm1:atr_bps'
        ]

    @staticmethod
    def get_strategy_algo_params() -> List[Dict[str, Any]]:
        return [
            {
                'key' : 'one_shot_only',
                'val' : True
            }
        ]